<?php

if ( !defined( 'ABSPATH' ) ) {
    exit;
}


if ( !class_exists( 'WTARS_Shipped_Premium_Shipping_Rates_Engine' ) ) {

    class WTARS_Shipped_Premium_Shipping_Rates_Engine {

        private $is_available_cache;
        private static $instance = null;

        private static function get_instance() {
            if ( null == self::$instance ) {
                self::$instance = new self( false );
            }
            return self::$instance;
        }

        public function __construct( $add_filter = false ) {

            if ( $add_filter ) {
                $this->is_available_cache = new WTARS_Shipped_Premium_Shipping_Rates_Cache();
                add_filter( 'wtars_shipped/process-data', array( $this, 'process_data' ), 2, 1 );
            }
        }

        public function process_data( $data ) {

            //shipping instance needs to be set
            if ( !isset( $data[ 'wc' ][ 'instance_id' ] ) ) {
                return $data;
            }

            $instance_id = $data[ 'wc' ][ 'instance_id' ];

            // check for cached or in memory data
            if ( isset( $data[ 'is_availables' ][ $instance_id ] ) ) {
                return $data;
            }


            //calculate shipping is_available
            $data = $this->is_available( $data, $instance_id );

            //store cache data
            if ( isset( $data[ 'is_availables' ][ $instance_id ] ) ) {
                $this->is_available_cache->store_cache( $data[ 'is_availables' ], $instance_id );
            } else {
                $this->is_available_cache->store_cache( array(), $instance_id );
            }

            return $data;
        }

        public static function calculate_costs( $rate_option, $instance_id, $data ) {

            $this_obj = self::get_instance();

            $max_args = $this_obj->get_max_args( $rate_option, $instance_id );

            $cost = $this_obj->calculate_base_costs( $rate_option, $instance_id, $data );

            if ( $this_obj->get_reached_limit( $cost, $max_args, $data ) ) {

                return $this_obj->get_cost_max( $cost, $max_args, $data );
            }

            $cost = $this_obj->calculate_extra_costs( $cost, $rate_option, $max_args, $instance_id, $data );

            return $this_obj->get_cost_max( $cost, $max_args, $data );
        }

        public static function can_apply_rate( $rate_option_head, $data ) {

            $instance_id = $rate_option_head[ 'instance_id' ];

            if ( 'if_others' == $rate_option_head[ 'apply_mode' ] && !isset( $data[ 'rates' ][ $instance_id ] ) ) {
                return false;
            }

            if ( 'if_no_others' == $rate_option_head[ 'apply_mode' ] && isset( $data[ 'rates' ][ $instance_id ] ) ) {
                return false;
            }

            if ( 'first' == $rate_option_head[ 'settings' ][ 'mode' ] && isset( $data[ 'rates' ][ $instance_id ] ) ) {
                return false;
            }


            return true;
        }

        public static function add_rate( $data, $in_rate, $rate_option_head ) {

            $instance_id = $rate_option_head[ 'instance_id' ];

            $rate = self::get_instance()->get_rate_before_add( $in_rate, $rate_option_head, $data );

            $s_mode = $rate_option_head[ 'settings' ][ 'mode' ];

            // add last valid rates data or only rate if valid
            if ( 'last' == $s_mode || 'only_this' == $rate_option_head[ 'apply_mode' ] ) {
                $data = self::get_instance()->force_rate( $data, $rate, $instance_id, $rate_option_head );
                return $data;
            }

            // add rate data with higher cost only
            if ( 'higher' == $s_mode ) {
                if ( self::get_instance()->get_is_higher( $data, $rate, $instance_id ) ) {
                    $data = self::get_instance()->force_rate( $data, $rate, $instance_id, $rate_option_head );
                }
                return $data;
            }

            // add rate data with lower cost only
            if ( 'lower' == $s_mode ) {
                if ( self::get_instance()->get_is_lower( $data, $rate, $instance_id ) ) {
                    $data = self::get_instance()->force_rate( $data, $rate, $instance_id, $rate_option_head );
                }
                return $data;
            }

            // add rate data as usual
            $data[ 'rates' ][ $instance_id ][ $rate_option_head[ 'rate_id' ] ] = $rate;

            // allows other plugins to modify rate data once added
            if ( has_filter( 'wtars_shipped/shipping-rates/added-rate' ) ) {
                if ( isset( $data[ 'rates' ][ $instance_id ][ $rate_option_head[ 'rate_id' ] ] ) ) {
                    $data[ 'rates' ][ $instance_id ][ $rate_option_head[ 'rate_id' ] ] = apply_filters( 'wtars_shipped/shipping-rates/added-rate', $data[ 'rates' ][ $instance_id ][ $rate_option_head[ 'rate_id' ] ], $instance_id, $rate_option_head );
                }
            }

            return $data;
        }

        private function calculate_base_costs( $rate_option, $instance_id, $data ) {

            $costs_args = $this->get_valid_costs_args( $this->get_costs_args_from_rate_option( $rate_option, 'base-costs', $instance_id ), $data );

            $costs = WTARS_Shipped_Cost_Types::calculate_costs( array(), $costs_args, $data );

            $merge_args = $this->get_merge_args( $rate_option, 'base-costs', $instance_id );

            $cost = WTARS_Shipped_Cost_Types::merge_costs( $costs, $merge_args );

            if ( has_filter( 'wtars_shipped/shipping-rates/calculated-base-cost' ) ) {
                $cost = apply_filters( 'wtars_shipped/shipping-rates/calculated-base-cost', $cost, $rate_option[ 'rate_id' ], $data );
            }

            return $cost;
        }

        private function calculate_extra_costs( $cost, $rate_option, $max_args, $instance_id, $data ) {

            $costs_args = $this->get_costs_args_from_rate_option( $rate_option, 'extra-costs', $instance_id );

            if ( !count( $costs_args ) ) {
                return $cost;
            }

            $merge_args = $this->get_merge_args( $rate_option, 'extra-costs', $instance_id );

            $extra_cost = array(
                'cost' => 0,
                'tax_total' => 0,
                'taxes' => array(),
            );

            foreach ( $costs_args as $cost_args ) {

                $costs = array();

                $grouped_data = $this->get_grouped_data( $data, $cost_args );

                //remove product args (products are already filtered)
                if ( isset( $cost_args[ 'products_args' ] ) ) {
                    unset( $cost_args[ 'products_args' ] );
                }

                // remove conditions
                if ( isset( $cost_args[ 'conditions' ] ) ) {
                    unset( $cost_args[ 'conditions' ] );
                }

                unset( $cost_args[ 'set_conditions' ] );


                $costs = $this->calculate_group_costs( $costs, $cost_args, $grouped_data );

                $new_cost = WTARS_Shipped_Cost_Types::merge_costs( $costs, $merge_args );

                $extra_cost = $this->combine_costs( $extra_cost, $new_cost );

                if ( $this->get_reached_limit( $this->combine_costs( $cost, $extra_cost ), $max_args, $data ) ) {
                    break;
                }
            }

            // allows other plugins to modify grouped package costs
            if ( has_filter( 'wtars_shipped/shipping-rates/calculated-extra-cost' ) ) {
                $extra_cost = apply_filters( 'wtars_shipped/shipping-rates/calculated-extra-cost', $extra_cost, $rate_option[ 'rate_id' ], $data );
            }

            return $this->combine_costs( $cost, $extra_cost );
        }

        private function calculate_group_costs( $costs, $cost_args, $grouped_data ) {

            foreach ( $grouped_data as $data ) {

                if ( !isset( $data[ 'wc' ][ 'package' ][ 'items' ] ) ) {
                    continue; //skip empty package items
                }

                if ( !count( $data[ 'wc' ][ 'package' ][ 'items' ] ) ) {
                    continue; //skip empty package items
                }

                $costs = WTARS_Shipped_Cost_Types::calculate_costs( $costs, array( $cost_args ), $data );
            }
            return $costs;
        }

        private function get_grouped_data( $data, $cost_args ) {

            $group_by = $cost_args[ 'group_by' ];

            $product_filters = isset( $cost_args[ 'products_args' ] ) ? $cost_args[ 'products_args' ] : array();

            $grd_data = WTARS_Shipped_Groups::get_packages( $data, $group_by, $product_filters );

            $grouped_data = array();

            foreach ( $grd_data as $key => $grp ) {

                if ( $this->get_cost_args_is_valid( $cost_args, $grp ) ) {

                    $grouped_data[ $key ] = $grp;
                }
            }

            return $grouped_data;
        }

        private function get_merge_args( $rate_option, $sub_module, $instance_id ) {
            return $merge_args = array(
                'rate_id' => $rate_option[ 'rate_id' ],
                'instance_id' => $instance_id,
                'module' => 'shipping-rates',
                'sub_module' => $sub_module,
            );
        }

        private function get_reached_limit( $cost, $max_args, $data ) {

            $cost_value = $cost[ 'cost' ];

            if ( 'yes' == $max_args[ 'inclusive_tax' ] ) {
                $cost_value = $cost[ 'cost' ] + $cost[ 'tax_total' ];
            }

            if ( $this->can_use_rate_limit( $max_args[ 'costs_limit' ] ) ) {

                $max_cost = $this->get_rate_limit( $max_args[ 'costs_limit' ], $data );

                return ($cost_value >= $max_cost);
            }

            if ( $this->can_use_global_limit( $max_args[ 'settings' ] ) ) {

                $max_cost = $this->get_global_limit( $max_args[ 'settings' ], $data );

                return ($cost_value >= $max_cost);
            }

            return false;
        }

        private function get_rate_before_add( $rate, $rate_option_head, $data ) {

            $instance_id = $rate_option_head[ 'instance_id' ];

            $max_args = $this->get_max_args( $rate_option_head, $instance_id );

            if ( $this->can_use_rate_limit( $max_args[ 'costs_limit' ] ) ) {

                $rate[ 'max' ] = $this->get_rate_limit( $max_args[ 'costs_limit' ], $data );
            }

            if ( $this->can_use_global_limit( $max_args[ 'settings' ] ) ) {

                $rate[ 'max' ] = $this->get_global_limit( $max_args[ 'settings' ], $data );
            }

            return $rate;
        }

        private function combine_costs( $prev_cost, $new_cost ) {

            if ( !isset( $prev_cost[ 'cost' ] ) ) {
                $prev_cost[ 'cost' ] = 0;
            }

            if ( !isset( $prev_cost[ 'tax_total' ] ) ) {
                $prev_cost[ 'tax_total' ] = 0;
            }

            $prev_cost[ 'cost' ] += $new_cost[ 'cost' ];

            $prev_cost[ 'tax_total' ] += $new_cost[ 'tax_total' ];

            foreach ( $new_cost[ 'taxes' ] as $key => $tax ) {

                if ( !isset( $prev_cost[ 'taxes' ][ $key ] ) ) {

                    $prev_cost[ 'taxes' ][ $key ] = 0;
                }

                $prev_cost[ 'taxes' ][ $key ] += $tax;
            }
            return $prev_cost;
        }

        private function get_costs_args_from_rate_option( $rate_option, $sub_module, $instance_id ) {

            $costs_args = array();

            $option_key = 'base_costs';

            if ( $sub_module == 'extra-costs' ) {
                $option_key = 'extra_costs';
            }

            if ( isset( $rate_option[ $option_key ] ) ) {

                $costs_args = $rate_option[ $option_key ];
            }

            $arg_keys = array_keys( $costs_args );

            $tax_type = $this->get_tax_type( $instance_id );

            foreach ( $arg_keys as $key ) {

                $costs_args[ $key ][ 'instance_id' ] = $instance_id;
                $costs_args[ $key ][ 'module' ] = 'shipping-rates';
                $costs_args[ $key ][ 'sub_module' ] = $sub_module;
                $costs_args[ $key ][ 'rate_id' ] = $rate_option[ 'rate_id' ];
                $costs_args[ $key ][ 'tax_type' ] = $tax_type;
                $costs_args[ $key ][ 'inclusive_tax' ] = $rate_option[ 'inclusive_tax' ];
            }

            return $costs_args;
        }

        private function get_valid_costs_args( $costs_args, $data ) {

            $args = array();
            foreach ( $costs_args as $key => $cost_args ) {

                if ( !isset( $cost_args[ 'conditions' ] ) ) {

                    $args[ $key ] = $cost_args;
                    continue;
                }

                if ( !WTARS_Shipped_Cost_Types::can_calculate_cost( $cost_args, $data ) ) {
                    continue; //do nothing
                }

                $args[ $key ] = $cost_args;

                unset( $args[ $key ][ 'set_conditions' ] );
                unset( $args[ $key ][ 'conditions' ] );
            }

            return $args;
        }

        private function get_cost_args_is_valid( $cost_args, $data ) {

            if ( !isset( $cost_args[ 'conditions' ] ) ) {
                return true;
            }

            return WTARS_Shipped_Cost_Types::can_calculate_cost( $cost_args, $data );
        }

        private function get_cost_max( $cost, $max_args, $data ) {

            $cost_value = $cost[ 'cost' ];

            if ( 'yes' == $max_args[ 'inclusive_tax' ] ) {
                $cost_value = $cost_value + $cost[ 'tax_total' ];
            }

            // adjust cost based on max cost by rate
            if ( $this->can_use_rate_limit( $max_args[ 'costs_limit' ] ) ) {

                $max_cost = $this->get_rate_limit( $max_args[ 'costs_limit' ], $data );

                if ( $cost_value > $max_cost ) {
                    $cost = $this->get_adjusted_cost( $cost, $max_cost, $max_args[ 'inclusive_tax' ] );
                }

                return $cost;
            }

            // adjust cost based on global max cost
            if ( $this->can_use_global_limit( $max_args[ 'settings' ] ) ) {

                $max_cost = $this->get_global_limit( $max_args[ 'settings' ], $data );

                if ( $cost_value > $max_cost ) {
                    $cost = $this->get_adjusted_cost( $cost, $max_cost, $max_args[ 'inclusive_tax' ] );
                }

                return $cost;
            }

            return $cost;
        }

        private function force_rate( $data, $rate, $instance_id, $rate_option_head ) {

            if ( isset( $data[ 'rates' ][ $instance_id ] ) ) {
                unset( $data[ 'rates' ][ $instance_id ] );
            }

            $data[ 'rates' ][ $instance_id ][ $rate_option_head[ 'rate_id' ] ] = $rate;

            return $data;
        }

        private function get_adjusted_cost( $cost, $new_cost, $inclusive_tax ) {

            foreach ( $cost[ 'taxes' ] as $key => $tax ) {

                $new_tax = ($new_cost / $cost[ 'cost' ]) * $tax;

                $cost[ 'taxes' ][ $key ] = WTARS_Shipped_Util::round_num( $new_tax, wc_get_price_decimals() );
            }

            $cost[ 'cost' ] = $new_cost;
            $cost[ 'tax_total' ] = array_sum( $cost[ 'taxes' ] );

            if ( 'yes' == $inclusive_tax ) {
                $cost[ 'cost' ] = $new_cost - $cost[ 'tax_total' ];
            }

            return $cost;
        }

        private function get_rate_limit( $limit_args, $data ) {

            if ( 'no' == $limit_args[ 'limit_type' ] ) {
                return 0;
            }

            $opt_limit = $limit_args[ 'limit' ];

            if ( !is_numeric( $opt_limit ) ) {
                $opt_limit = 0;
            }

            if ( 'fixed_limit' == $limit_args[ 'limit_type' ] ) {
                return $opt_limit;
            }

            if ( 'per_limit' == $limit_args[ 'limit_type' ] ) {

                $args = array(
                    'module' => 'shipping-rates',
                    'sub_module' => 'rates',
                    'instance_id' => $data[ 'wc' ][ 'instance_id' ],
                    'option_id' => $limit_args[ 'based_on' ]
                );

                return $this->get_per_limit( $opt_limit, $args, $data );
            }

            return 0;
        }

        private function get_global_limit( $settings, $data ) {

            if ( 'no' == $settings[ 'limit_type' ] ) {
                return 0;
            }

            $opt_limit = $settings[ 'limit' ];
            if ( !is_numeric( $opt_limit ) ) {
                $opt_limit = 0;
            }

            if ( 'fixed_limit' == $settings[ 'limit_type' ] ) {

                return $opt_limit;
            }

            if ( 'per_limit' == $settings[ 'limit_type' ] ) {

                $args = array(
                    'module' => 'shipping-rates',
                    'sub_module' => 'rates',
                    'instance_id' => $data[ 'wc' ][ 'instance_id' ],
                    'option_id' => $settings[ 'based_on' ]
                );

                return $this->get_per_limit( $opt_limit, $args, $data );
            }

            return 0;
        }

        private function get_per_limit( $opt_limit, $args, $data ) {

            $totals = WTARS_Shipped_Total_Types::get_totals( $args, $data );

            if ( $totals > 0 ) {
                return (($opt_limit / 100) * $totals);
            }
            return 0;
        }

        private function get_is_higher( $data, $rate, $instance_id ) {

            $value = 0;

            if ( !isset( $data[ 'rates' ][ $instance_id ] ) ) {
                return true;
            }

            foreach ( $data[ 'rates' ][ $instance_id ] as $prev_rate ) {

                if ( $prev_rate[ 'cost' ] > $value ) {
                    $value = $prev_rate[ 'cost' ];
                }
            }

            return ($rate[ 'cost' ] > $value);
        }

        private function get_is_lower( $data, $rate, $instance_id ) {

            $value = 0;

            if ( !isset( $data[ 'rates' ][ $instance_id ] ) ) {
                return true;
            }

            foreach ( $data[ 'rates' ][ $instance_id ] as $prev_rate ) {

                if ( $value < $prev_rate[ 'cost' ] ) {
                    $value = $prev_rate[ 'cost' ];
                }
            }

            return ($rate[ 'cost' ] < $value);
        }

        private function can_use_rate_limit( $limit_args ) {

            if ( !isset( $limit_args[ 'limit' ] ) || !is_numeric( $limit_args[ 'limit' ] ) ) {
                return false;
            }

            if ( $limit_args[ 'limit' ] <= 0 ) {
                return false;
            }

            return ('no' != $limit_args[ 'limit_type' ]);
        }

        private function can_use_global_limit( $settings ) {

            if ( !isset( $settings[ 'limit' ] ) || !is_numeric( $settings[ 'limit' ] ) ) {
                return false;
            }

            if ( $settings[ 'limit' ] <= 0 ) {
                return false;
            }

            return ( 'no' != $settings[ 'limit_type' ] );
        }

        private function get_tax_type( $instance_id ) {

            $default = array(
                'tax_type' => 'ship_rates',
            );

            $shipping = WTARS_Shipped::get_option( 'shipping_rate_settings', $instance_id, $default );

            return $shipping[ 'tax_type' ];
        }

        private function get_max_args( $rate_option, $instance_id ) {

            $max_args = array(
                'costs_limit' => $rate_option[ 'costs_limit' ],
                'settings' => $this->get_rates_settings( $instance_id ),
                'inclusive_tax' => $rate_option[ 'inclusive_tax' ],
            );
            return $max_args;
        }

        private function is_available( $data, $instance_id ) {

            //prepare conditions args
            $method_conditions = $this->get_method_conditions( $instance_id );

            $method_conditions[ 'instance_id' ] = $instance_id;
            $method_conditions[ 'module' ] = 'method';
            $method_conditions[ 'sub_module' ] = 'method';

            //validate method conditions
            $is_available = WTARS_Shipped_Condition_Types::validate_conditions( $method_conditions, $data );
            $data[ 'is_availables' ][ $instance_id ] = $is_available;

            return $data;
        }

        private function get_method_conditions( $instance_id ) {

            $default = array(
                'match_mode' => 'match_all',
            );

            return WTARS_Shipped::get_option( 'method_conditions', $instance_id, $default );
        }

        private function get_rates_settings( $instance_id ) {

            $default = array(
                'mode' => 'no',
            );

            return WTARS_Shipped::get_option( 'shipping_rates_settings', $instance_id, $default );
        }

    }

}
